#!/usr/bin/perl -w
#
# Copyright (C) 2001,2002 asdfGroup.com Inc. All rights reserved.
#
# This library is a commercial product and it is subject to the license
# agreement that can be found at the following URL:
#   http://www.asdfGroup.com/license.html
# Please contact our staff if you will have any questions:
#   mailto:staff@asdfGroup.com
#

=head1 NAME

B<POWER::LOG> -  

=head1 SYNOPSIS

 use POWER::LOG;

 $LOG = POWER::LOG->new();
 $LOG->config({
    OUTPUT  => \*STDERR,
    PREFIX  => "%D %T %P::%F() : ",
    LEVEL   => "WARN",
    });
 # OR
 $LOG = $POWER::LOG::DEFAULT;
 $LOG->config({
    OUTPUT  => \*STDERR,
    LEVEL   => "WARN",
    });
 
 $LOG->ERR($message);
 $LOG->WARN($message);
 $LOG->INFO($message);
 $LOG->DEBUG($message);
 $LOG->DUMP($message);

=head1 DESCRIPTION

      .
       .
     {LEVEL}   
.

=head1 METHODS

=over 4

=cut

package POWER::LOG;

use POWER::lib qw(time);

use strict;

our $VERSION = "2.22";
our $DEFAULT =
    POWER::LOG->new(PREFIX=>'%S %L %_', OUTPUT=>\*STDERR, LEVEL=>"ERR");
our $LEVEL;

=item B<new>( [\%config] )

         .
 $config,   ,   config().

 new()     POWER::LOG,   
.

=cut

sub new {
    my $this  = shift;
    my $class = ref($this) || $this ;
    my $self  = { CONFIG => {}, };
    bless $self, $class ;
    $self->config($this->config()) if ref($this);
    $self->config(@_);
    return $self;
}

=item B<config>( [%config] )

  .
%config -    .
    ,   B<CONFIG>.

      .

     2.22 config()   %config 
\%config.

=cut

sub config {
    my ($self, $config) = (shift, ref($_[0]) eq "HASH" ? $_[0] : {@_});
eval 'use Carp; carp \'Use config(%data) instead of config(\%data)\'' if ref($_[0]);
    UpdateHash($self->{CONFIG}, $config);
    # TODO ,  ref($self) ne "POWER::LOG"
    bless $self, ref($self->{CONFIG}{OUTPUT}) ne "GLOB" 
	? "POWER::LOG::NULL" : "POWER::LOG";
    @{$self}{"ERR","WARN","INFO","DEBUG","DUMP"} = 
        $_ eq "ERR"  ? (1,0,0,0,0) : $_ eq "WARN"  ? (1,1,0,0,0) : 
        $_ eq "INFO" ? (1,1,1,0,0) : $_ eq "DEBUG" ? (1,1,1,1,0) :
        $_ eq "DUMP" ? (1,1,1,1,1) : (0,0,0,0,0) for $self->{CONFIG}{LEVEL};
    my @pfx = split /(\%.)/, $self->{CONFIG}{PREFIX};
    my %need;
    for (@pfx) {
        /^\%\%$/ && do{$_='"%"'}
     || /^\%L/ && do{$_='$LEVEL'}
     || /^\%S/ && do{$_='$data{S}'; $need{Sec}      = 1 }
     || /^\%D/ && do{$_='$data{D}'; $need{DateTime} = 1 }
     || /^\%T/ && do{$_='$data{T}'; $need{DateTime} = 1 }
     || /^\%P/ && do{$_='$data{P}'; $need{Caller}   = 1 }
     || /^\%F/ && do{$_='$data{F}'; $need{Caller}   = 1; $need{Func}    = 1 }
     || /^\%_/ && do{$_='$data{_}'; $need{Stack}    = 1 }
     ||           do{$_="\"\Q$_\E\""};
    }
    my $code = '{ my ($was, $now, $t, @t, %data) = (0); sub {';
    $code .= '
        $t = time();
        $data{S} = sprintf \'%d.%05d\', $t, ($t-int($t))*100000;' if $need{Sec};
    $code .= '
        if (($now = CORE::time()) != $was) {
            ($was, @t) = ($now, localtime());
            $data{D} = sprintf \'%02d/%02d/%04d\', $t[3], $t[4]+1, $t[5]+1900;
            $data{T} = sprintf \'%02d:%02d:%02d\', $t[2], $t[1], $t[0];
        }' if $need{DateTime};
    $code .= '
        $data{P} = caller(1);' if $need{Caller};
    $code .= '
        $data{F} = substr((caller(2))[3], length($data{P})+2);' if $need{Func};
    $code .= '
        $data{_}=3; 1 while caller($data{_}++); $data{_} = " " x ($data{_}-3);'
        if $need{Stack};
    $code .= 'return join "", ('.join(',',grep {!/^\"\"$/} @pfx).'); }}';
    $self->{prefix} = eval $code;
    return %{$self->{CONFIG}};
}

=item B<ERR>( $message[, @list] )

     . [ ]

  @list,   printf($prefix.$message."\n", @list),
  print($prefix, $message, "\n").   @list 
  ,         
  @list     . :

 #   Bin2HexDump  {LEVEL}  "DUMP"
 $LOG->DUMP("Received packet:\n%s", sub { Bin2HexDump($pkt) });

=item B<WARN>( $message[, @list] )

     . [  ]

=item B<INFO>( $message[, @list] )

   . [   /  ]

=item B<DEBUG>( $message[, @list] )

    . [   ]

=item B<DUMP>( $message[, @list] )

       .

=cut

{ local $LEVEL;
sub ERR     { @_==2 ? goto &_log : goto &_logf if $_[0]->{$LEVEL="ERR"}   }
sub WARN    { @_==2 ? goto &_log : goto &_logf if $_[0]->{$LEVEL="WARN"}  }
sub INFO    { @_==2 ? goto &_log : goto &_logf if $_[0]->{$LEVEL="INFO"}  }
sub DEBUG   { @_==2 ? goto &_log : goto &_logf if $_[0]->{$LEVEL="DEBUG"} }
sub DUMP    { @_==2 ? goto &_log : goto &_logf if $_[0]->{$LEVEL="DUMP"}  }
sub _log    { print  {$_[0]->{CONFIG}{OUTPUT}} $_[0]->{prefix}->(), $_[1], "\n" }
sub _logf   {
    printf {$_[0]->{CONFIG}{OUTPUT}} '%s'.$_[1]."\n", $_[0]->{prefix}->(), 
	map {ref $_ eq "CODE" ? $_->() : $_} @_[2..$#_]
}
}

package POWER::LOG::NULL;
our @ISA = qw(POWER::LOG);
sub ERR {}
sub WARN {}
sub INFO {}
sub DEBUG {}
sub DUMP {}

=back

=head1 CONFIG

=over 4

=item B<{OUTPUT}>

 ,     .   ,
    .

=item B<{PREFIX}>

,        .
 {PREFIX}     ,   
   .     printf(),  
,     :

 %L - LEVEL,    
 %S -      
 %D -    :  "//"
 %T -    : "::"
 %P -   : "main"  "POWER::NB::IO"
 %F -   :   "my_func_name"
 %_ - X ,  X -  
 %% -  "%"

 : "%D %T %P::%F() "

=item B<{LEVEL}>

 {LEVEL}        : "ERR","WARN","INFO",
"DEBUG","DUMP"     .   
,       {LEVEL}. , 
{LEVEL}  "INFO"     "ERR", "WARN"  "INFO".

=back

=head1 SPEED

 POWER::LOG   {OUTPUT}   
    10000    .

       -
 : %L, %P, %%; : %S, %D, %T; : %F, %_.

   ({OUTPUT}    ),  
 .

  $message    @list,   printf() 
print()     .

=head1 HISTORY


 2.22	      31 14:18:35 EET 2003
	- bugfix: config()    
	- incompatible changes: config()    
	  ,   ( config() -    
	    ,   warn)
	- incompatible changes: config()    
	  ,  
	- added printf() feature
	- added code-ref feature in printf() params
 2.21	      23 17:23:16 EEST 2002
	-      {OUTPUT}
 2.20         16 11:25:08 EEST 2002
        -  
        -  
        - minor bug fixes
 2.12          9 04:19:21 EEST 2002
        -       10 
 2.11         27 16:36:06 EET 2002
        -  : %L, %S, %_
        - DEFAULT   {OUTPUT=>\*STDERR,LEVEL=>"ERR"}
 2.10         25 10:41:50 EET 2002
        -  $POWER::LOG::DEFAULT
 2.01
 2.00         10 05:42:11 EET 2002
        -    
 1.01         28 21:28:52 EET 2001
        -   /
 1.00         28 16:48:16 EET 2001
        -     ( print)

=head1 AUTHOR

 Alex Efros <powerman@asdfGroup.com>

 Copyright (C) 2001,2002 asdfGroup.com Inc. All rights reserved.

 This library is a commercial product and it is subject to the license
 agreement that can be found at the following URL:
    http://www.asdfGroup.com/license.html
 Please contact our staff if you will have any questions:
    mailto:staff@asdfGroup.com

=cut

1;
